"""
Aqua vuln-list to NVD CVE converter

This module fetches the vulnerability data from aquasec vuln-list repo and stores them in NVD CVE 1.1 json format.
"""

import os
from datetime import datetime
from zipfile import ZipFile

import httpx
import orjson

from vdb.lib import CustomNamedTemporaryFile, config
from vdb.lib.nvd import NvdSource
from vdb.lib.utils import (
    compress_str,
    convert_score_severity,
    get_cvss3_from_vector,
    get_default_cve_data,
)

# Size of the stream to read and write to the file
DOWNLOAD_CHUNK_SIZE = 4096


class AquaSource(NvdSource):
    """Aqua CVE source"""

    def download_all(self):
        """Download all cve data"""
        # For performance do not retain the whole data in-memory
        # See: https://github.com/AppThreat/vulnerability-db/issues/27
        data_list = []
        self.fetch(config.VULN_LIST_URL)
        return data_list

    def download_recent(self):
        pass

    def _process_zip(self, zname):
        with ZipFile(zname, "r") as zipfp:
            for zf in zipfp.namelist():
                if self.is_supported_source(zf):
                    with zipfp.open(zf) as jsonf:
                        cve_data = jsonf.read()
                        try:
                            json_data = orjson.loads(cve_data)
                            ret_data = self.convert(json_data)
                            self.store(ret_data)
                        except Exception:
                            pass

    def fetch(self, url):
        # Check if there is an existing cached zip file we could use
        cached_zip = os.path.join(config.CACHE_DIR, "vuln-list.zip")
        if os.path.exists(cached_zip):
            return self._process_zip(cached_zip)
        with CustomNamedTemporaryFile() as tf:
            try:
                client = httpx.Client(http2=True, follow_redirects=True, timeout=180)
                with client.stream("GET", url) as r:
                    for chunk in r.iter_bytes(chunk_size=DOWNLOAD_CHUNK_SIZE):
                        tf.write(chunk)
                    tf.flush()
                    return self._process_zip(tf.name)
            except Exception:
                return []

    def convert(self, cve_data):
        if cve_data.get("vulnStatus"):
            return self.nvd_api_to_vuln(cve_data)
        if cve_data.get("updateinfo_id"):
            return self.alsa_to_vuln(cve_data)
        if cve_data.get("id", "").startswith("ALAS"):
            return self.alas_rlsa_to_vuln(cve_data, "amazon")
        if cve_data.get("id", "").startswith("RLSA"):
            return self.alas_rlsa_to_vuln(cve_data, "rocky")
        if cve_data.get("Candidate"):
            return self.ubuntu_to_vuln(cve_data)
        if cve_data.get("affected_release"):
            return self.redhat_to_vuln(cve_data)
        if cve_data.get("name", "").startswith("AVG"):
            return self.arch_to_vuln(cve_data)
        if cve_data.get("Tracking"):
            return self.suse_to_vuln(cve_data)
        if cve_data.get("os_version"):
            return self.photon_to_vuln(cve_data)
        if cve_data.get("Annotations") and cve_data.get("Header"):
            return self.debian_to_vuln(cve_data)
        if cve_data.get("secfixes"):
            return self.wolfi_to_vuln(cve_data)
        if cve_data.get("id", "").startswith("CVE-") and cve_data.get("state"):
            return self.alpine_to_vuln(cve_data)
        return []

    @staticmethod
    def is_supported_source(zfname):
        for distro in ("alpine-unfixed",):
            if distro in zfname:
                return True
        for distro in (
            "alpine",
            "cwe",
            "ghsa",
            "go",
            "osv",
            "redhat-cpe",
            "kevc",
            "oval",
            "glad",
            "mariner",
        ):
            if distro in zfname:
                return False
        nvd_start_year = config.NVD_START_YEAR
        for year in range(1999, nvd_start_year):
            if f"CVE-{year}-" in zfname:
                return False
        if zfname.endswith(".json"):
            return True
        return False

    @staticmethod
    def alsa_to_vuln(cve_data):
        """AlmaLinux"""
        ret_data = []
        if cve_data.get("type") != "security":
            return ret_data
        packages = cve_data.get("pkglist", {}).get("packages", [])
        if not packages or not len(packages) > 0:
            return ret_data
        cve_id = cve_data.get("updateinfo_id")
        cwe_id = ""
        cve_references = cve_data.get("references", [])
        references = []
        for aref in cve_references:
            references.append(
                {"name": aref.get("title", "id"), "url": aref.get("href")}
            )
        references = orjson.dumps(references, option=orjson.OPT_NAIVE_UTC)
        if isinstance(references, bytes):
            references = references.decode("utf-8", "ignore")
        description = cve_data.get("description", "")
        if not description and cve_data.get("title"):
            description = f"""# {cve_data.get("summary")}
{cve_data.get("title")}
{cve_data.get("solution")}
            """
        assigner = cve_data.get("fromstr", "")
        severity = config.THREAT_TO_SEVERITY[cve_data.get("severity").lower()]
        score, severity, vector_string, attack_complexity = get_default_cve_data(
            severity
        )
        exploitability_score = score
        user_interaction = "REQUIRED"
        vendor = "almalinux"
        pkg_name = packages[0].get("name")
        version_start_including = ""
        version_end_including = ""
        version_start_excluding = ""
        fix_version_end_including = ""
        fix_version_start_excluding = ""
        fix_version_end_excluding = ""
        fix_version_start_including = packages[0].get("version").split(":")[-1]
        version_end_excluding = packages[0].get("version").split(":")[-1]
        published_date = ""
        if cve_data.get("issued_date", {}).get("$date"):
            try:
                published_date = datetime.fromtimestamp(
                    int(cve_data.get("issued_date", {}).get("$date")) / 1000
                )
            except Exception:
                pass
        last_modified_date = ""
        if cve_data.get("updated_date", {}).get("$date"):
            try:
                last_modified_date = datetime.fromtimestamp(
                    int(cve_data.get("updated_date", {}).get("$date")) / 1000
                )
            except Exception:
                pass
        if pkg_name and fix_version_start_including:
            tdata = config.CVE_TPL % dict(
                cve_id=cve_id,
                cwe_id=cwe_id,
                assigner=assigner,
                references=references,
                description="",
                vectorString=vector_string,
                vendor="rpm",
                product=f"{vendor}/{pkg_name}",
                version="*",
                edition="*",
                version_start_including=version_start_including,
                version_end_including=version_end_including,
                version_start_excluding=version_start_excluding,
                version_end_excluding=version_end_excluding,
                fix_version_start_including=fix_version_start_including,
                fix_version_end_including=fix_version_end_including,
                fix_version_start_excluding=fix_version_start_excluding,
                fix_version_end_excluding=fix_version_end_excluding,
                severity=severity,
                attackComplexity=attack_complexity,
                score=score,
                userInteraction=user_interaction,
                exploitabilityScore=exploitability_score,
                publishedDate=published_date,
                lastModifiedDate=last_modified_date,
            )
            try:
                vuln = NvdSource.convert_vuln(orjson.loads(tdata))
                vuln.description = compress_str(description)
                ret_data.append(vuln)
            except Exception:
                pass
        return ret_data

    @staticmethod
    def alas_rlsa_to_vuln(cve_data, vendor):
        """Amazon Linux"""
        ret_data = []
        packages = cve_data.get("packages", [])
        cve_id = cve_data.get("id")
        if not packages or cve_id in ("CVE-PENDING",) or not len(packages) > 0:
            return ret_data
        cwe_id = ""
        cve_references = cve_data.get("references", [])
        references = []
        for aref in cve_references:
            references.append({"name": aref.get("id"), "url": aref.get("href")})
        references = orjson.dumps(references, option=orjson.OPT_NAIVE_UTC)
        if isinstance(references, bytes):
            references = references.decode("utf-8", "ignore")
        if not cve_id.startswith("CVE") and cve_data.get("cveids"):
            cve_id = cve_data.get("cveids")[0]
        description = cve_data.get("description", "")
        assigner = vendor
        severity = config.THREAT_TO_SEVERITY[cve_data.get("severity").lower()]
        score, severity, vector_string, attack_complexity = get_default_cve_data(
            severity
        )
        exploitability_score = score
        user_interaction = "REQUIRED"
        published_date = cve_data.get("issued", {}).get("date")
        last_modified_date = cve_data.get("updated", {}).get("date")
        done_pkgs = {}
        for apkg in packages:
            version = apkg.get("version")
            # Remove epoch
            if ":" in version:
                version = version.split(":")[-1]
            pkg_key = f"""{apkg.get("name")}:{version}"""
            if done_pkgs.get(pkg_key):
                continue
            pkg_name = apkg.get("name")
            version_start_including = ""
            version_end_including = ""
            version_start_excluding = ""
            fix_version_end_including = ""
            fix_version_start_excluding = ""
            fix_version_end_excluding = ""
            fix_version_start_including = version
            version_end_excluding = version

            if pkg_name and fix_version_start_including:
                tdata = config.CVE_TPL % dict(
                    cve_id=cve_id,
                    cwe_id=cwe_id,
                    assigner=assigner,
                    references=references,
                    description="",
                    vectorString=vector_string,
                    vendor="rpm",
                    product=f"{vendor}/{pkg_name}",
                    version="*",
                    edition="*",
                    version_start_including=version_start_including,
                    version_end_including=version_end_including,
                    version_start_excluding=version_start_excluding,
                    version_end_excluding=version_end_excluding,
                    fix_version_start_including=fix_version_start_including,
                    fix_version_end_including=fix_version_end_including,
                    fix_version_start_excluding=fix_version_start_excluding,
                    fix_version_end_excluding=fix_version_end_excluding,
                    severity=severity,
                    attackComplexity=attack_complexity,
                    score=score,
                    userInteraction=user_interaction,
                    exploitabilityScore=exploitability_score,
                    publishedDate=published_date,
                    lastModifiedDate=last_modified_date,
                )
                try:
                    vuln = NvdSource.convert_vuln(orjson.loads(tdata))
                    vuln.description = compress_str(description)
                    ret_data.append(vuln)
                    done_pkgs[pkg_key] = True
                except Exception:
                    pass
        return ret_data

    @staticmethod
    def ubuntu_to_vuln(cve_data):
        """Ubuntu Linux"""
        ret_data = []
        packages = cve_data.get("Patches", {})
        if not packages or not len(packages) > 0:
            return ret_data
        cve_id = cve_data.get("Candidate")
        cwe_id = ""
        cve_references = cve_data.get("References", [])
        references = []
        for aref in cve_references:
            references.append({"name": aref, "url": aref})
        references = orjson.dumps(references, option=orjson.OPT_NAIVE_UTC)
        if isinstance(references, bytes):
            references = references.decode("utf-8", "ignore")
        description = cve_data.get("Description")
        if "** DISPUTED **" in description or "** REJECT **" in description:
            return ret_data
        if not description and cve_data.get("UbuntuDescription"):
            description = cve_data.get("UbuntuDescription")
        assigner = "canonical"
        vendor = "ubuntu"
        severity = config.THREAT_TO_SEVERITY[cve_data.get("Priority").lower()]
        score, severity, vector_string, attack_complexity = get_default_cve_data(
            severity
        )
        exploitability_score = score
        user_interaction = "REQUIRED"
        published_date = cve_data.get("PublicDate")
        last_modified_date = cve_data.get("PublicDate")
        for pkg_name, distro_obj in packages.items():
            for distro_name, status_obj in distro_obj.items():
                fix_note = status_obj.get("Note")
                # DNE - does not exist
                if status_obj.get("Status") in ("DNE",):
                    continue
                version_start_including = ""
                version_end_including = ""
                version_start_excluding = ""
                version_end_excluding = ""
                fix_version_end_including = ""
                fix_version_start_excluding = ""
                fix_version_end_excluding = ""
                fix_version_start_including = ""
                # Remove epoch
                if ":" in fix_note:
                    fix_note = fix_note.split(":")[-1]
                # Released CVEs have fixes
                if (
                    status_obj.get("Status") in ("not-affected", "released")
                    and " " not in fix_note
                    and "CVE" not in fix_note
                ):
                    fix_version_start_including = fix_note
                    version_end_excluding = fix_note
                # Handle CVEs that are deferred
                # Let's include the vulnerabilities that did not get a fix
                if (
                    status_obj.get("Status") in ("deferred", "needed")
                    and " " not in fix_note
                ):
                    version_end_including = config.PLACEHOLDER_FIX_VERSION
                if status_obj.get("Status") in ("ignored", "needs-triage"):
                    version_end_including = config.PLACEHOLDER_FIX_VERSION
                if pkg_name and (fix_version_start_including or version_end_including):
                    for full_pkg_name in (f"{distro_name}/{pkg_name}",):
                        tdata = config.CVE_TPL % dict(
                            cve_id=cve_id,
                            cwe_id=cwe_id,
                            assigner=assigner,
                            references=references,
                            description="",
                            vectorString=vector_string,
                            vendor="deb",
                            product=f"{vendor}/{full_pkg_name}",
                            version="*",
                            edition=distro_name,
                            version_start_including=version_start_including,
                            version_end_including=version_end_including,
                            version_start_excluding=version_start_excluding,
                            version_end_excluding=version_end_excluding,
                            fix_version_start_including=fix_version_start_including,
                            fix_version_end_including=fix_version_end_including,
                            fix_version_start_excluding=fix_version_start_excluding,
                            fix_version_end_excluding=fix_version_end_excluding,
                            severity=severity,
                            attackComplexity=attack_complexity,
                            score=score,
                            userInteraction=user_interaction,
                            exploitabilityScore=exploitability_score,
                            publishedDate=published_date,
                            lastModifiedDate=last_modified_date,
                        )
                        try:
                            vuln = NvdSource.convert_vuln(orjson.loads(tdata))
                            vuln.description = compress_str(description)
                            ret_data.append(vuln)
                        except Exception:
                            pass
        return ret_data

    @staticmethod
    def redhat_to_vuln(cve_data):
        """RedHat Linux"""
        ret_data = []
        cvss3 = cve_data.get("cvss3", {})
        if cvss3.get("status") is not None and cvss3.get("status") != "verified":
            return ret_data
        packages = cve_data.get("affected_release", [])
        if not packages or not len(packages) > 0:
            return ret_data
        cve_id = cve_data.get("name")
        cwe_id = cve_data.get("cwe")
        cve_references = cve_data.get("references", "")
        references = []
        for aref in cve_references:
            for bref in aref.split("\n"):
                references.append({"name": bref, "url": bref})
        references = orjson.dumps(references, option=orjson.OPT_NAIVE_UTC)
        if isinstance(references, bytes):
            references = references.decode("utf-8", "ignore")
        description = "\n".join(cve_data.get("details", []))
        assigner = "redhat"
        vector_string = cvss3.get("cvss3_scoring_vector")
        score = cvss3.get("cvss3_base_score")
        if score:
            try:
                score = float(score)
            except Exception:
                pass
        severity = config.THREAT_TO_SEVERITY[cve_data.get("threat_severity").lower()]
        cvss3_obj = get_cvss3_from_vector(vector_string)
        exploitability_score = ""
        attack_complexity = ""
        user_interaction = "REQUIRED"
        if cvss3_obj:
            exploitability_score = cvss3_obj.get("temporalScore")
            attack_complexity = cvss3_obj.get("attackComplexity")
            score = cvss3_obj.get("baseScore")
            user_interaction = cvss3_obj.get("userInteraction")
        published_date = cve_data.get("public_date")
        last_modified_date = cve_data.get("public_date")
        done_pkgs = {}
        for arelease in packages:
            pkg_key = arelease.get("package")
            if done_pkgs.get(pkg_key):
                continue
            tmp_a = pkg_key.split(":" if ":" in pkg_key else "-")
            if len(tmp_a) < 2:
                continue
            pkg_name = tmp_a[0]
            if pkg_name.endswith("-0"):
                pkg_name = pkg_name[0:-2]
            if ":" in pkg_key:
                version = pkg_key.split(":")[-1]
            else:
                version = pkg_key.replace(pkg_name + "-", "")
            # Remove epoch
            if ":" in version:
                version = version.split(":")[-1]
            edition = "*"
            cpe = arelease.get("cpe", "")
            if cpe and cpe.startswith("cpe:"):
                tmpc = cpe.split(":")
                if len(tmpc) > 2:
                    edition = f"{tmpc[-2]}-{tmpc[-1]}"
                    if edition.startswith("-el"):
                        edition = edition.replace("-el", "enterprise_linux-")
            version_start_including = ""
            version_end_including = version
            version_start_excluding = ""
            version_end_excluding = ""
            fix_version_end_including = ""
            fix_version_start_excluding = version
            fix_version_end_excluding = ""
            fix_version_start_including = ""
            if pkg_name and version:
                tdata = config.CVE_TPL % dict(
                    cve_id=cve_id,
                    cwe_id=cwe_id,
                    assigner=assigner,
                    references=references,
                    description="",
                    vectorString=vector_string,
                    vendor="rpm",
                    product=f"redhat/{pkg_name}",
                    version="*",
                    edition=edition,
                    version_start_including=version_start_including,
                    version_end_including=version_end_including,
                    version_start_excluding=version_start_excluding,
                    version_end_excluding=version_end_excluding,
                    fix_version_start_including=fix_version_start_including,
                    fix_version_end_including=fix_version_end_including,
                    fix_version_start_excluding=fix_version_start_excluding,
                    fix_version_end_excluding=fix_version_end_excluding,
                    severity=severity,
                    attackComplexity=attack_complexity,
                    score=score,
                    userInteraction=user_interaction,
                    exploitabilityScore=exploitability_score,
                    publishedDate=published_date,
                    lastModifiedDate=last_modified_date,
                )
                try:
                    vuln = NvdSource.convert_vuln(orjson.loads(tdata))
                    vuln.description = compress_str(description)
                    ret_data.append(vuln)
                    done_pkgs[pkg_key] = True
                except Exception:
                    pass
        return ret_data

    @staticmethod
    def arch_to_vuln(cve_data):
        """Arch Linux"""
        ret_data = []
        packages = cve_data.get("packages", [])
        if not packages or not len(packages) > 0:
            return ret_data
        cve_id = cve_data.get("name")
        cwe_id = ""
        references = []
        if cve_data.get("issues") and len(cve_data.get("issues")) > 0:
            cve_id = cve_data.get("issues")[0]
        description = cve_data.get("type", "")
        assigner = "archlinux"
        severity = config.THREAT_TO_SEVERITY[cve_data.get("severity").lower()]
        score, severity, vector_string, attack_complexity = get_default_cve_data(
            severity
        )
        published_date = ""
        last_modified_date = ""
        for pkg_name in packages:
            version_start_including = ""
            version_end_including = cve_data.get("affected")
            version_start_excluding = ""
            version_end_excluding = cve_data.get("fixed")
            fix_version_end_including = ""
            fix_version_start_excluding = ""
            fix_version_end_excluding = ""
            fix_version_start_including = cve_data.get("fixed")
            if pkg_name and version_end_including:
                tdata = config.CVE_TPL % dict(
                    cve_id=cve_id,
                    cwe_id=cwe_id,
                    assigner=assigner,
                    references=references,
                    description="",
                    vectorString=vector_string,
                    vendor="alpm",
                    product=f"arch/{pkg_name}",
                    version="*",
                    edition="*",
                    version_start_including=version_start_including,
                    version_end_including=version_end_including,
                    version_start_excluding=version_start_excluding,
                    version_end_excluding=version_end_excluding,
                    fix_version_start_including=fix_version_start_including,
                    fix_version_end_including=fix_version_end_including,
                    fix_version_start_excluding=fix_version_start_excluding,
                    fix_version_end_excluding=fix_version_end_excluding,
                    severity=severity,
                    attackComplexity=attack_complexity,
                    score=score,
                    userInteraction="REQUIRED",
                    exploitabilityScore=score,
                    publishedDate=published_date,
                    lastModifiedDate=last_modified_date,
                )
                try:
                    vuln = NvdSource.convert_vuln(orjson.loads(tdata))
                    vuln.description = compress_str(description)
                    ret_data.append(vuln)
                except Exception:
                    pass
        return ret_data

    @staticmethod
    def product_ref_to_name_version(product_ref):
        product_parts = product_ref.split("-")
        name_parts = []
        version_parts = []
        for i, part in enumerate(product_parts):
            if part[0].isdigit() and "bit" not in part and "kb" not in part:
                version_parts = product_parts[i:]
                name_parts = product_parts[:i]
                break
        return "-".join(name_parts), "-".join(version_parts)

    def suse_to_vuln(self, cve_data):
        """Suse Linux"""
        ret_data = []
        packages = cve_data.get("ProductTree", {}).get("Relationships", [])
        if not packages or not len(packages) > 0:
            return ret_data
        severity = ""
        # Package name has to be extracted from the title :(
        published_date = cve_data.get("Tracking", {}).get("InitialReleaseDate", "")
        last_modified_date = cve_data.get("Tracking", {}).get("CurrentReleaseDate", "")
        if cve_data.get("Vulnerabilities"):
            for avuln in cve_data.get("Vulnerabilities"):
                cve_id = avuln.get("CVE")
                description = avuln.get("Description")
                threats = avuln.get("Threats")
                if threats and len(threats) > 0:
                    severity = config.THREAT_TO_SEVERITY[
                        threats[0].get("Severity").lower()
                    ]
                cwe_id = ""
                cve_references = avuln.get("References", [])
                references = []
                for aref in cve_references:
                    references.append(
                        {"name": aref.get("Description", "id"), "url": aref.get("URL")}
                    )
                references = orjson.dumps(references, option=orjson.OPT_NAIVE_UTC)
                if isinstance(references, bytes):
                    references = references.decode("utf-8", "ignore")
                assigner = "suse"
                (
                    score,
                    severity,
                    vector_string,
                    attack_complexity,
                ) = get_default_cve_data(severity)
                done_pkgs = {}
                for pref in packages:
                    pkg_key = pref.get("ProductReference")
                    if done_pkgs.get(pkg_key):
                        continue
                    pkg_name, version = self.product_ref_to_name_version(pkg_key)
                    version_start_including = ""
                    version_end_including = ""
                    version_start_excluding = ""
                    version_end_excluding = version
                    fix_version_end_including = ""
                    fix_version_start_excluding = ""
                    fix_version_end_excluding = ""
                    fix_version_start_including = version
                    if pkg_name and version:
                        tdata = config.CVE_TPL % dict(
                            cve_id=cve_id,
                            cwe_id=cwe_id,
                            assigner=assigner,
                            references=references,
                            description="",
                            vectorString=vector_string,
                            vendor="rpm",
                            product=f"suse/{pkg_name}",
                            version="*",
                            edition="*",
                            version_start_including=version_start_including,
                            version_end_including=version_end_including,
                            version_start_excluding=version_start_excluding,
                            version_end_excluding=version_end_excluding,
                            fix_version_start_including=fix_version_start_including,
                            fix_version_end_including=fix_version_end_including,
                            fix_version_start_excluding=fix_version_start_excluding,
                            fix_version_end_excluding=fix_version_end_excluding,
                            severity=severity,
                            attackComplexity=attack_complexity,
                            score=score,
                            userInteraction="REQUIRED",
                            exploitabilityScore=score,
                            publishedDate=published_date,
                            lastModifiedDate=last_modified_date,
                        )
                        try:
                            vuln = NvdSource.convert_vuln(orjson.loads(tdata))
                            vuln.description = compress_str(description)
                            ret_data.append(vuln)
                            done_pkgs[pkg_key] = True
                        except Exception:
                            pass
        return ret_data

    @staticmethod
    def photon_to_vuln(cve_data):
        """Photon Linux"""
        ret_data = []
        cve_id = cve_data.get("cve_id")
        pkg_name = cve_data.get("pkg")
        cwe_id = ""
        references = []
        description = f"""Summary
{cve_data.get("aff_ver")}
        """
        assigner = "vmware"
        score = cve_data.get("cve_score")
        severity = convert_score_severity(score)
        newscore, severity, vector_string, attack_complexity = get_default_cve_data(
            severity
        )
        if not score and newscore:
            score = newscore
        published_date = ""
        last_modified_date = ""
        version_start_including = ""
        version_end_including = ""
        version_start_excluding = ""
        version_end_excluding = cve_data.get("res_ver").split(":")[-1]
        fix_version_end_including = ""
        fix_version_start_excluding = ""
        fix_version_end_excluding = ""
        fix_version_start_including = cve_data.get("res_ver").split(":")[-1]
        distro_name = "*"
        if cve_data.get("os_version"):
            distro_name = f"""photon-{cve_data.get("os_version")}"""
            pkg_name = f"{distro_name}/{pkg_name}"
        if pkg_name and version_end_excluding:
            tdata = config.CVE_TPL % dict(
                cve_id=cve_id,
                cwe_id=cwe_id,
                assigner=assigner,
                references=references,
                description="",
                vectorString=vector_string,
                vendor="rpm",
                product=f"photon/{pkg_name}",
                version="*",
                edition=distro_name,
                version_start_including=version_start_including,
                version_end_including=version_end_including,
                version_start_excluding=version_start_excluding,
                version_end_excluding=version_end_excluding,
                fix_version_start_including=fix_version_start_including,
                fix_version_end_including=fix_version_end_including,
                fix_version_start_excluding=fix_version_start_excluding,
                fix_version_end_excluding=fix_version_end_excluding,
                severity=severity,
                attackComplexity=attack_complexity,
                score=score,
                userInteraction="REQUIRED",
                exploitabilityScore=score,
                publishedDate=published_date,
                lastModifiedDate=last_modified_date,
            )
            try:
                vuln = NvdSource.convert_vuln(orjson.loads(tdata))
                vuln.description = compress_str(description)
                ret_data.append(vuln)
            except Exception:
                pass
        return ret_data

    @staticmethod
    def debian_to_vuln(cve_data):
        """Debian Linux"""
        ret_data = []
        header = cve_data.get("Header")
        annotations = cve_data.get("Annotations")
        cve_id = header.get("ID")
        cwe_id = ""
        references = []
        description = header.get("Description", "").replace("(", "").replace(")", "")
        # Try harder to retain all description
        for ann in annotations:
            if ann.get("Original", "").startswith("NOTE:"):
                description = f"{description}\n{ann.get('Original')}"
        if "** DISPUTED **" in description or "** REJECT **" in description:
            return ret_data
        assigner = "debian"
        vendor = "debian"
        for ann in annotations:
            if ann.get("Type") == "RESERVED" or ann.get("Original") == "RESERVED":
                continue
            # Try to dealias
            if (
                not cve_id.startswith("CVE")
                and ann.get("Type") == "xref"
                and ann.get("Bugs")
            ):
                aliases_block = f"""
## Related CVE(s)
{", ".join(ann.get("Bugs"))}
            """
                description += aliases_block
                for bug in ann.get("Bugs"):
                    if bug.startswith("CVE"):
                        cve_id = bug
                        break
            kind = ann.get("Kind")
            if kind not in (
                "fixed",
                "unfixed",
                "no-dsa",
                "end-of-life",
                "ignored",
                "not-affected",
                "postponed",
            ):
                continue
            pkg_name = ann.get("Package")
            if not pkg_name:
                continue
            distro_name = ann.get("Release")
            if distro_name:
                pkg_name = f"{distro_name}/{pkg_name}"
            version = ann.get("Version", config.PLACEHOLDER_FIX_VERSION)
            # We need to track not-affected entries with a special exclude version
            if kind == "not-affected":
                version = config.PLACEHOLDER_EXCLUDE_VERSION
            # Remove epoch
            if ":" in version:
                version = version.split(":")[-1]
            severity = "MEDIUM"
            # Try harder to set LOW priority
            if kind == "postponed" or (
                ann.get("Description")
                and "minor issue" in ann.get("Description").lower()
            ):
                severity = "LOW"
            if ann.get("Severity"):
                severity = config.THREAT_TO_SEVERITY.get(ann.get("Severity"), "MEDIUM")
            score, severity, vector_string, attack_complexity = get_default_cve_data(
                severity
            )
            published_date = ""
            last_modified_date = ""
            version_start_including = ""
            version_end_including = ""
            version_start_excluding = ""
            version_end_excluding = version
            fix_version_end_including = ""
            fix_version_start_excluding = ""
            fix_version_end_excluding = ""
            fix_version_start_including = version
            if pkg_name and version_end_excluding:
                tdata = config.CVE_TPL % dict(
                    cve_id=cve_id,
                    cwe_id=cwe_id,
                    assigner=assigner,
                    references=references,
                    description="",
                    vectorString=vector_string,
                    vendor="deb",
                    product=f"{vendor}/{pkg_name}",
                    version="*",
                    edition=distro_name if distro_name else "*",
                    version_start_including=version_start_including,
                    version_end_including=version_end_including,
                    version_start_excluding=version_start_excluding,
                    version_end_excluding=version_end_excluding,
                    fix_version_start_including=fix_version_start_including,
                    fix_version_end_including=fix_version_end_including,
                    fix_version_start_excluding=fix_version_start_excluding,
                    fix_version_end_excluding=fix_version_end_excluding,
                    severity=severity,
                    attackComplexity=attack_complexity,
                    score=score,
                    userInteraction="REQUIRED",
                    exploitabilityScore=score,
                    publishedDate=published_date,
                    lastModifiedDate=last_modified_date,
                )
                try:
                    vuln = NvdSource.convert_vuln(orjson.loads(tdata))
                    vuln.description = compress_str(description)
                    ret_data.append(vuln)
                except Exception:
                    pass
        return ret_data

    @staticmethod
    def wolfi_to_vuln(cve_data):
        """Wolfi OS and Chainguard"""
        ret_data = []
        cwe_id = ""
        references = []
        assigner = cve_data.get("reponame")
        # No severity of any kind in the advisories
        severity = "LOW"
        score, severity, vector_string, attack_complexity = get_default_cve_data(
            severity
        )
        published_date = ""
        last_modified_date = ""
        pkg_name = cve_data.get("name")
        for fix_version_start_including, cve_list in cve_data.get("secfixes").items():
            for cve_id in cve_list:
                version_start_including = ""
                version_end_including = (
                    "*" if fix_version_start_including == "0" else ""
                )
                version_start_excluding = ""
                version_end_excluding = (
                    fix_version_start_including
                    if fix_version_start_including != "0"
                    else ""
                )
                fix_version_end_including = ""
                fix_version_start_excluding = ""
                fix_version_end_excluding = ""
                tdata = config.CVE_TPL % dict(
                    cve_id=cve_id,
                    cwe_id=cwe_id,
                    assigner=assigner,
                    references=references,
                    description="",
                    vectorString=vector_string,
                    vendor="apk",
                    product=f"{assigner}/{pkg_name}",
                    version="*",
                    edition="*",
                    version_start_including=version_start_including,
                    version_end_including=version_end_including,
                    version_start_excluding=version_start_excluding,
                    version_end_excluding=version_end_excluding,
                    fix_version_start_including=fix_version_start_including,
                    fix_version_end_including=fix_version_end_including,
                    fix_version_start_excluding=fix_version_start_excluding,
                    fix_version_end_excluding=fix_version_end_excluding,
                    severity=severity,
                    attackComplexity=attack_complexity,
                    score=score,
                    userInteraction="REQUIRED",
                    exploitabilityScore=score,
                    publishedDate=published_date,
                    lastModifiedDate=last_modified_date,
                )
                try:
                    vuln = NvdSource.convert_vuln(orjson.loads(tdata))
                    vuln.description = f"""URL Prefix: {cve_data.get("urlprefix")}. Affected arch: {", ".join(cve_data.get("archs"))}"""
                    ret_data.append(vuln)
                except Exception:
                    pass
        return ret_data

    @staticmethod
    def alpine_to_vuln(cve_data):
        """Alpine"""
        ret_data = []
        assigner = "alpine"
        # No severity of any kind in the advisories
        severity = "LOW"
        score, severity, vector_string, attack_complexity = get_default_cve_data(
            severity
        )
        published_date = ""
        last_modified_date = ""
        cve_id = cve_data.get("id")
        for astate in cve_data.get("state", []):
            if astate.get("fixed"):
                continue
            pkg_name = astate.get("packageName")
            unfixed_version = astate.get("packageVersion")
            edition = f"alpine-{astate.get('repo').split('-')[0]}"
            tdata = config.CVE_TPL % dict(
                cve_id=cve_id,
                cwe_id="",
                assigner=assigner,
                references=[],
                description="",
                vectorString=vector_string,
                vendor="apk",
                product=f"{assigner}/{edition}/{pkg_name}",
                version=unfixed_version,
                edition=edition,
                version_start_including=unfixed_version,
                version_end_including=unfixed_version,
                version_start_excluding="",
                version_end_excluding="",
                fix_version_start_including="",
                fix_version_end_including="",
                fix_version_start_excluding="",
                fix_version_end_excluding="",
                severity=severity,
                attackComplexity=attack_complexity,
                score=score,
                userInteraction="NONE",
                exploitabilityScore=score,
                publishedDate=published_date,
                lastModifiedDate=last_modified_date,
            )
            try:
                vuln = NvdSource.convert_vuln(orjson.loads(tdata))
                vuln.description = f"{cve_id} is unfixed in {edition}. Visit https://security.alpinelinux.org/vuln/{cve_id} for more details."
                ret_data.append(vuln)
            except Exception:
                pass
        return ret_data

    @staticmethod
    def nvd_api_to_vuln(cve_data):
        """NVD API"""
        v = NvdSource.convert_api_vuln(cve_data)
        return [v] if v else []



# pylint: disable=C0115, C0103, C0301

from __future__ import annotations

from enum import Enum
from typing import Annotated, Any, Dict, List, Optional, Type, Union

from pydantic import AnyUrl, AwareDatetime, BaseModel, ConfigDict, Field, NaiveDatetime, RootModel

from vdb.lib.cve_model import cvss_v2, cvss_v3


class UriType(RootModel[AnyUrl]):
    root: Annotated[
        AnyUrl,
        Field(
            description="A universal resource identifier (URI), according to [RFC 3986](https://tools.ietf.org/html/rfc3986)."
        ),
    ]


class UuidType(RootModel[str]):
    root: Annotated[
        str,
        Field(
            description="A version 4 (random) universally unique identifier (UUID) as defined by [RFC 4122](https://tools.ietf.org/html/rfc4122#section-4.1.3).",
            pattern="^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-4[0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$",
        ),
    ]


class ReferenceTags(Enum):
    broken_link = "broken-link"
    customer_entitlement = "customer-entitlement"
    exploit = "exploit"
    government_resource = "government-resource"
    issue_tracking = "issue-tracking"
    mailing_list = "mailing-list"
    mitigation = "mitigation"
    not_applicable = "not-applicable"
    patch = "patch"
    permissions_required = "permissions-required"
    media_coverage = "media-coverage"
    product = "product"
    related = "related"
    release_notes = "release-notes"
    signature = "signature"
    technical_description = "technical-description"
    third_party_advisory = "third-party-advisory"
    vendor_advisory = "vendor-advisory"
    vdb_entry = "vdb-entry"


# Removed patten validation to support a wide range of CVE id
class CveId(RootModel[str]):
    root: Annotated[str, Field(min_length=4, max_length=128)]


class OrgId(RootModel[UuidType]):
    root: Annotated[
        UuidType,
        Field(
            description="A UUID for an organization participating in the CVE program. This UUID can be used to lookup the organization record in the user registry service."
        ),
    ]


class UserId(RootModel[UuidType]):
    root: Annotated[
        UuidType,
        Field(
            description="A UUID for a user participating in the CVE program. This UUID can be used to lookup the user record in the user registry service."
        ),
    ]


class ShortName(RootModel[str]):
    root: Annotated[
        str,
        Field(
            description="A 2-32 character name that can be used to complement an organization's UUID.",
            max_length=32,
            min_length=2,
        ),
    ]


class Version(RootModel[str]):
    root: Annotated[
        str,
        Field(
            description="A single version of a product, as expressed in its own version numbering scheme.",
            max_length=1024,
            min_length=1,
        ),
    ]


class Status(Enum):
    affected = "affected"
    unaffected = "unaffected"
    unknown = "unknown"


# Removed pattern validation since it wasn't working for go full names
class Cpe(RootModel[str]):
    root: Annotated[
        str,
        Field(
            description="Common Platform Enumeration (CPE) Name in either 2.2 or 2.3 format",
            max_length=2048,
            min_length=1,
            title="CPE Name",
        ),
    ]


class Module(RootModel[str]):
    root: Annotated[
        str,
        Field(
            description="Name of the affected component, feature, module, sub-component, sub-product, API, command, utility, program, or functionality (optional).",
            min_length=1,
        ),
    ]


class ProgramFile(RootModel[str]):
    root: Annotated[
        str,
        Field(
            description="Name or path or location of the affected source code file.",
            max_length=1024,
            min_length=1,
        ),
    ]


class ProgramRoutine(BaseModel):
    name: Annotated[
        str,
        Field(
            description="Name of the affected source code file, function, method, subroutine, or procedure.",
            min_length=1,
        ),
    ]


class Platform(RootModel[str]):
    root: Annotated[
        str,
        Field(
            examples=[
                "iOS",
                "Android",
                "Windows",
                "macOS",
                "x86",
                "ARM",
                "64 bit",
                "Big Endian",
                "iPad",
                "Chromebook",
                "Docker",
                "Model T",
            ],
            max_length=1024,
        ),
    ]


class Change(BaseModel):
    at: Annotated[
        Version, Field(description="The version at which a status change occurs.")
    ]
    status: Annotated[
        Status,
        Field(description="The new status in the range starting at the given version."),
    ]


class Versions(BaseModel):
    version: Annotated[
        Version,
        Field(
            description="The single version being described, or the version at the start of the range. By convention, typically 0 denotes the earliest possible version."
        ),
    ]
    status: Annotated[
        Status,
        Field(
            description="The vulnerability status for the version or range of versions. For a range, the status may be refined by the 'changes' list."
        ),
    ]
    versionType: Annotated[
        Optional[str],
        Field(
            None,
            description="The version numbering system used for specifying the range. This defines the exact semantics of the comparison (less-than) operation on versions, which is required to understand the range itself. 'Custom' indicates that the version type is unspecified and should be avoided whenever possible. It is included primarily for use in conversion of older data files.",
            examples=["custom", "git", "maven", "python", "rpm", "semver"],
            max_length=128,
            min_length=1,
        ),
    ]
    lessThan: Annotated[
        Optional[Version],
        Field(
            None,
            description="The non-inclusive upper limit of the range. This is the least version NOT in the range. The usual version syntax is expanded to allow a pattern to end in an asterisk `(*)`, indicating an arbitrarily large number in the version ordering. For example, `{version: 1.0 lessThan: 1.*}` would describe the entire 1.X branch for most range kinds, and `{version: 2.0, lessThan: *}` describes all versions starting at 2.0, including 3.0, 5.1, and so on. Only one of lessThan and lessThanOrEqual should be specified.",
        ),
    ]
    lessThanOrEqual: Annotated[
        Optional[Version],
        Field(
            None,
            description="The inclusive upper limit of the range. This is the greatest version contained in the range. Only one of lessThan and lessThanOrEqual should be specified. For example, `{version: 1.0, lessThanOrEqual: 1.3}` covers all versions from 1.0 up to and including 1.3.",
        ),
    ]
    changes: Annotated[
        Optional[List[Change]],
        Field(
            None,
            description="A list of status changes that take place during the range. The array should be sorted in increasing order by the 'at' field, according to the versionType, but clients must re-sort the list themselves rather than assume it is sorted.",
            min_length=1,
        ),
    ]


class Versions1(BaseModel):
    version: Annotated[
        Version,
        Field(
            description="The single version being described, or the version at the start of the range. By convention, typically 0 denotes the earliest possible version."
        ),
    ]
    status: Annotated[
        Status,
        Field(
            description="The vulnerability status for the version or range of versions. For a range, the status may be refined by the 'changes' list."
        ),
    ]
    versionType: Annotated[
        str,
        Field(
            description="The version numbering system used for specifying the range. This defines the exact semantics of the comparison (less-than) operation on versions, which is required to understand the range itself. 'Custom' indicates that the version type is unspecified and should be avoided whenever possible. It is included primarily for use in conversion of older data files.",
            examples=["custom", "git", "maven", "python", "rpm", "semver"],
            max_length=128,
            min_length=1,
        ),
    ]
    lessThan: Annotated[
        Version,
        Field(
            description="The non-inclusive upper limit of the range. This is the least version NOT in the range. The usual version syntax is expanded to allow a pattern to end in an asterisk `(*)`, indicating an arbitrarily large number in the version ordering. For example, `{version: 1.0 lessThan: 1.*}` would describe the entire 1.X branch for most range kinds, and `{version: 2.0, lessThan: *}` describes all versions starting at 2.0, including 3.0, 5.1, and so on. Only one of lessThan and lessThanOrEqual should be specified."
        ),
    ]
    lessThanOrEqual: Annotated[
        Optional[Version],
        Field(
            None,
            description="The inclusive upper limit of the range. This is the greatest version contained in the range. Only one of lessThan and lessThanOrEqual should be specified. For example, `{version: 1.0, lessThanOrEqual: 1.3}` covers all versions from 1.0 up to and including 1.3.",
        ),
    ]
    changes: Annotated[
        Optional[List[Change]],
        Field(
            None,
            description="A list of status changes that take place during the range. The array should be sorted in increasing order by the 'at' field, according to the versionType, but clients must re-sort the list themselves rather than assume it is sorted.",
            min_length=1,
        ),
    ]


class Versions2(BaseModel):
    version: Annotated[
        Version,
        Field(
            description="The single version being described, or the version at the start of the range. By convention, typically 0 denotes the earliest possible version."
        ),
    ]
    status: Annotated[
        Status,
        Field(
            description="The vulnerability status for the version or range of versions. For a range, the status may be refined by the 'changes' list."
        ),
    ]
    versionType: Annotated[
        str,
        Field(
            description="The version numbering system used for specifying the range. This defines the exact semantics of the comparison (less-than) operation on versions, which is required to understand the range itself. 'Custom' indicates that the version type is unspecified and should be avoided whenever possible. It is included primarily for use in conversion of older data files.",
            examples=["custom", "git", "maven", "python", "rpm", "semver"],
            max_length=128,
            min_length=1,
        ),
    ]
    lessThan: Annotated[
        Optional[Version],
        Field(
            None,
            description="The non-inclusive upper limit of the range. This is the least version NOT in the range. The usual version syntax is expanded to allow a pattern to end in an asterisk `(*)`, indicating an arbitrarily large number in the version ordering. For example, `{version: 1.0 lessThan: 1.*}` would describe the entire 1.X branch for most range kinds, and `{version: 2.0, lessThan: *}` describes all versions starting at 2.0, including 3.0, 5.1, and so on. Only one of lessThan and lessThanOrEqual should be specified.",
        ),
    ]
    lessThanOrEqual: Annotated[
        Version,
        Field(
            description="The inclusive upper limit of the range. This is the greatest version contained in the range. Only one of lessThan and lessThanOrEqual should be specified. For example, `{version: 1.0, lessThanOrEqual: 1.3}` covers all versions from 1.0 up to and including 1.3."
        ),
    ]
    changes: Annotated[
        Optional[List[Change]],
        Field(
            None,
            description="A list of status changes that take place during the range. The array should be sorted in increasing order by the 'at' field, according to the versionType, but clients must re-sort the list themselves rather than assume it is sorted.",
            min_length=1,
        ),
    ]


class Product(BaseModel):
    vendor: Annotated[
        Optional[str],
        Field(
            None,
            description="Name of the organization, project, community, individual, or user that created or maintains this product or hosted service. Can be 'N/A' if none of those apply. When collectionURL and packageName are used, this field may optionally represent the user or account within the package collection associated with the package.",
            max_length=512,
            min_length=1,
        ),
    ]
    product: Annotated[
        Optional[str],
        Field(
            None,
            description="Name of the affected product.",
            max_length=2048,
            min_length=1,
        ),
    ]
    collectionURL: Annotated[
        Optional[UriType],
        Field(
            None,
            description="URL identifying a package collection (determines the meaning of packageName).",
            examples=[
                "https://access.redhat.com/downloads/content/package-browser",
                "https://addons.mozilla.org",
                "https://addons.thunderbird.net",
                "https://anaconda.org/anaconda/repo",
                "https://app.vagrantup.com/boxes/search",
                "https://apps.apple.com",
                "https://archlinux.org/packages",
                "https://atmospherejs.meteor.com",
                "https://atom.io/packages",
                "https://bitbucket.org",
                "https://bower.io",
                "https://brew.sh/",
                "https://chocolatey.org/packages",
                "https://chrome.google.com/webstore",
                "https://clojars.org",
                "https://cocoapods.org",
                "https://code.dlang.org",
                "https://conan.io/center",
                "https://cpan.org/modules",
                "https://cran.r-project.org",
                "https://crates.io",
                "https://ctan.org/pkg",
                "https://drupal.org",
                "https://exchange.adobe.com",
                "https://forge.puppet.com/modules",
                "https://github.com",
                "https://gitlab.com/explore",
                "https://golang.org/pkg",
                "https://guix.gnu.org/packages",
                "https://hackage.haskell.org",
                "https://helm.sh",
                "https://hub.docker.com",
                "https://juliahub.com",
                "https://lib.haxe.org",
                "https://luarocks.org",
                "https://marketplace.visualstudio.com",
                "https://melpa.org",
                "https://microsoft.com/en-us/store/apps",
                "https://nimble.directory",
                "https://nuget.org/packages",
                "https://opam.ocaml.org/packages",
                "https://openwrt.org/packages/index",
                "https://package.elm-lang.org",
                "https://packagecontrol.io",
                "https://packages.debian.org",
                "https://packages.gentoo.org",
                "https://packagist.org",
                "https://pear.php.net/packages.php",
                "https://pecl.php.net",
                "https://platformio.org/lib",
                "https://play.google.com/store",
                "https://plugins.gradle.org",
                "https://projects.eclipse.org",
                "https://pub.dev",
                "https://pypi.python.org",
                "https://registry.npmjs.org",
                "https://registry.terraform.io",
                "https://repo.hex.pm",
                "https://repo.maven.apache.org/maven2",
                "https://rubygems.org",
                "https://search.nixos.org/packages",
                "https://sourceforge.net",
                "https://wordpress.org/plugins",
            ],
        ),
    ]
    packageName: Annotated[
        Optional[str],
        Field(
            None,
            description="Name or identifier of the affected software package as used in the package collection.",
            max_length=2048,
            min_length=1,
        ),
    ]
    cpes: Annotated[
        Optional[List[Cpe]],
        Field(
            None,
            description='Affected products defined by CPE. This is an array of CPE values (vulnerable and not), we use an array so that we can make multiple statements about the same version and they are separate (if we used a JSON object we\'d essentially be keying on the CPE name and they would have to overlap). Also, this allows things like cveDataVersion or cveDescription to be applied directly to the product entry. This also allows more complex statements such as "Product X between versions 10.2 and 10.8" to be put in a machine-readable format. As well since multiple statements can be used multiple branches of the same product can be defined here.',
        ),
    ]
    modules: Annotated[
        Optional[List[Module]],
        Field(
            None,
            description="A list of the affected components, features, modules, sub-components, sub-products, APIs, commands, utilities, programs, or functionalities (optional).",
        ),
    ]
    programFiles: Annotated[
        Optional[List[ProgramFile]],
        Field(None, description="A list of the affected source code files (optional)."),
    ]
    programRoutines: Annotated[
        Optional[List[ProgramRoutine]],
        Field(
            None,
            description="A list of the affected source code functions, methods, subroutines, or procedures (optional).",
        ),
    ]
    platforms: Annotated[
        Optional[List[Platform]],
        Field(
            None,
            description="List of specific platforms if the vulnerability is only relevant in the context of these platforms (optional). Platforms may include execution environments, operating systems, virtualization technologies, hardware models, or computing architectures. The lack of this field or an empty array implies that the other fields are applicable to all relevant platforms.",
            min_length=1,
            title="Platforms",
        ),
    ]
    repo: Annotated[
        Optional[UriType],
        Field(
            None,
            description="The URL of the source code repository, for informational purposes and/or to resolve git hash version ranges.",
        ),
    ]
    defaultStatus: Annotated[
        Optional[Status],
        Field(
            None,
            description="The default status for versions that are not otherwise listed in the versions list. If not specified, defaultStatus defaults to 'unknown'. Versions or defaultStatus may be omitted, but not both.",
        ),
    ]
    versions: Annotated[
        Optional[List[Union[Versions, Union[Versions1, Versions2]]]],
        Field(
            None,
            description="Set of product versions or version ranges related to the vulnerability. The versions satisfy the CNA Rules [8.1.2 requirement](https://cve.mitre.org/cve/cna/rules.html#section_8-1_cve_entry_information_requirements). Versions or defaultStatus may be omitted, but not both.",
            min_length=1,
        ),
    ]


class DataType(Enum):
    CVE_RECORD = "CVE_RECORD"


class DataVersion(Enum):
    field_5_0 = "5.0"


class State(Enum):
    PUBLISHED = "PUBLISHED"


class CveMetadataPublished(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    cveId: Annotated[
        CveId, Field(description="The CVE identifier that this record pertains to.")
    ]
    assignerOrgId: Annotated[
        OrgId,
        Field(
            description="The UUID for the organization to which the CVE ID was originally assigned. This UUID can be used to lookup the organization record in the user registry service."
        ),
    ]
    assignerShortName: Annotated[
        Optional[ShortName],
        Field(
            None,
            description="The short name for the organization to which the CVE ID was originally assigned.",
        ),
    ]
    requesterUserId: Annotated[
        Optional[UserId],
        Field(None, description="The user that requested the CVE identifier."),
    ]
    dateUpdated: Annotated[
        Optional[AwareDatetime | NaiveDatetime],
        Field(None, description="The date/time the record was last updated."),
    ]
    serial: Annotated[
        Optional[int],
        Field(
            None,
            description="The system of record causes this to start at 1, and increment by 1 each time a submission from a data provider changes this CVE Record. The incremented value moves to the Rejected schema upon a PUBLISHED->REJECTED transition, and moves to the Published schema upon a REJECTED->PUBLISHED transition.",
            ge=1,
        ),
    ]
    dateReserved: Annotated[
        Optional[AwareDatetime | NaiveDatetime],
        Field(
            None,
            description="The date/time this CVE ID was reserved in the CVE automation workgroup services system. Disclaimer: This date reflects when the CVE ID was reserved, and does not necessarily indicate when this vulnerability was discovered, shared with the affected vendor, publicly disclosed, or updated in CVE.",
        ),
    ]
    datePublished: Annotated[
        Optional[AwareDatetime | NaiveDatetime],
        Field(
            None,
            description="The date/time the CVE Record was first published in the CVE List.",
        ),
    ]
    state: Annotated[State, Field(description="State of CVE - PUBLISHED, REJECTED.")]


class State1(Enum):
    REJECTED = "REJECTED"


class CveMetadataRejected(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    cveId: Annotated[
        CveId, Field(description="The CVE identifier that this record pertains to.")
    ]
    assignerOrgId: Annotated[
        OrgId,
        Field(
            description="The UUID for the organization to which the CVE ID was originally assigned."
        ),
    ]
    assignerShortName: Annotated[
        Optional[ShortName],
        Field(
            None,
            description="The short name for the organization to which the CVE ID was originally assigned.",
        ),
    ]
    serial: Annotated[
        Optional[int],
        Field(
            None,
            description="The system of record causes this to start at 1, and increment by 1 each time a submission from a data provider changes this CVE Record. The incremented value moves to the Rejected schema upon a PUBLISHED->REJECTED transition, and moves to the Published schema upon a REJECTED->PUBLISHED transition.",
            ge=1,
        ),
    ]
    dateUpdated: Annotated[
        Optional[AwareDatetime | NaiveDatetime],
        Field(None, description="The date/time the record was last updated."),
    ]
    datePublished: Annotated[
        Optional[AwareDatetime | NaiveDatetime],
        Field(
            None,
            description="The date/time the CVE Record was first published in the CVE List.",
        ),
    ]
    dateRejected: Annotated[
        Optional[AwareDatetime | NaiveDatetime],
        Field(None, description="The date/time the CVE ID was rejected."),
    ]
    state: Annotated[State1, Field(description="State of CVE - PUBLISHED, REJECTED.")]
    dateReserved: Annotated[
        Optional[AwareDatetime | NaiveDatetime],
        Field(
            None,
            description="The date/time this CVE ID was reserved in the CVE automation workgroup services system. Disclaimer: This date reflects when the CVE ID was reserved, and does not necessarily indicate when this vulnerability was discovered, shared with the affected vendor, publicly disclosed, or updated in CVE.",
        ),
    ]


class ProviderMetadata(BaseModel):
    orgId: Annotated[
        OrgId, Field(description="The container provider's organizational UUID.")
    ]
    shortName: Annotated[
        Optional[ShortName],
        Field(None, description="The container provider's organizational short name."),
    ]
    dateUpdated: Annotated[
        Optional[AwareDatetime | NaiveDatetime],
        Field(
            None,
            description="Timestamp to be set by the system of record at time of submission. If dateUpdated is provided to the system of record it will be replaced by the current timestamp at the time of submission.",
        ),
    ]


class Affected(RootModel[List[Product]]):
    root: Annotated[
        List[Product], Field(description="List of affected products.", min_length=1)
    ]


class SupportingMediaItem(BaseModel):
    type: Annotated[
        str,
        Field(
            description="RFC2046 compliant IANA Media type for eg., text/markdown, text/html.",
            examples=[
                "text/markdown",
                "text/html",
                "image/png",
                "image/svg",
                "audio/mp3",
            ],
            max_length=256,
            min_length=1,
            title="Media type",
        ),
    ]
    base64: Annotated[
        Optional[bool],
        Field(
            False,
            description="If true then the value field contains the media data encoded in base64. If false then the value field contains the UTF-8 media content.",
            title="Encoding",
        ),
    ]
    # Removed max_length
    value: Annotated[
        str,
        Field(
            description="Supporting media content. If base64 is true, this field stores base64 encoded data.",
            min_length=1,
        ),
    ]


class Other(BaseModel):
    type: Annotated[
        str,
        Field(
            description="Name of the non-standard impact metrics format used.",
            max_length=128,
            min_length=1,
        ),
    ]
    content: Annotated[
        Dict[str, Any],
        Field(description="JSON object not covered by another metrics format."),
    ]


class Source(BaseModel):
    pass


class Language(RootModel[str]):
    root: Annotated[
        str,
        Field(
            description="BCP 47 language code, language-region.",
            pattern="^[A-Za-z]{2,4}([_-][A-Za-z]{4})?([_-]([A-Za-z]{2}|[0-9]{3}))?$",
        ),
    ]


class EnglishLanguage(RootModel[str]):
    root: Annotated[
        str,
        Field(
            description="BCP 47 language code, language-region, required to be English.",
            pattern="^en([_-][A-Za-z]{4})?([_-]([A-Za-z]{2}|[0-9]{3}))?$",
        ),
    ]


class TaxonomyRelation(BaseModel):
    taxonomyId: Annotated[
        str,
        Field(
            description="Identifier of the item in the taxonomy.  Used as the subject of the relationship.",
            max_length=2048,
            min_length=1,
        ),
    ]
    relationshipName: Annotated[
        str,
        Field(
            description="A description of the relationship.",
            max_length=128,
            min_length=1,
        ),
    ]
    relationshipValue: Annotated[
        str,
        Field(
            description="The target of the relationship.  Can be the CVE ID or another taxonomy identifier.",
            max_length=2048,
            min_length=1,
        ),
    ]


class TaxonomyMapping(BaseModel):
    taxonomyName: Annotated[
        str,
        Field(description="The name of the taxonomy.", max_length=128, min_length=1),
    ]
    taxonomyVersion: Annotated[
        Optional[str],
        Field(
            None,
            description="The version of taxonomy the identifiers come from.",
            max_length=128,
            min_length=1,
        ),
    ]
    taxonomyRelations: Annotated[
        List[TaxonomyRelation], Field(description="", min_length=1)
    ]


class TaxonomyMappings(RootModel[List[TaxonomyMapping]]):
    root: Annotated[
        List[TaxonomyMapping],
        Field(
            description="List of taxonomy items related to the vulnerability.",
            min_length=1,
        ),
    ]


class TagExtension(RootModel[str]):
    root: Annotated[str, Field(max_length=128, min_length=2, pattern="^x_.*$")]


class CnaTags(Enum):
    unsupported_when_assigned = "unsupported-when-assigned"
    exclusively_hosted_service = "exclusively-hosted-service"
    disputed = "disputed"


class AdpTags(Enum):
    disputed = "disputed"


class Reference(BaseModel):
    url: Annotated[
        UriType,
        Field(
            description="The uniform resource locator (URL), according to [RFC 3986](https://tools.ietf.org/html/rfc3986#section-1.1.3), that can be used to retrieve the referenced resource."
        ),
    ]
    name: Annotated[
        Optional[str],
        Field(
            None,
            description="User created name for the reference, often the title of the page.",
            max_length=512,
            min_length=1,
        ),
    ]
    tags: Annotated[
        Optional[List[Union[TagExtension, ReferenceTags]]],
        Field(
            None,
            description="An array of one or more tags that describe the resource referenced by 'url'.",
            min_length=1,
        ),
    ]


class Description(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    lang: Language
    value: Annotated[
        str, Field(description="Plain text description.", min_length=1, max_length=4096)
    ]
    supportingMedia: Annotated[
        Optional[List[SupportingMediaItem]],
        Field(
            None,
            description="Supporting media data for the description such as markdown, diagrams, .. (optional). Similar to RFC 2397 each media object has three main parts: media type, media data value, and an optional boolean flag to indicate if the media data is base64 encoded.",
            min_length=1,
            title="Supporting media",
        ),
    ]


class EnglishLanguageDescription(BaseModel):
    lang: EnglishLanguage


class Descriptions(RootModel[List[Description]]):
    root: Annotated[
        List[Description],
        Field(
            description="A list of multi-lingual descriptions of the vulnerability. E.g., [PROBLEMTYPE] in [COMPONENT] in [VENDOR] [PRODUCT] [VERSION] on [PLATFORMS] allows [ATTACKER] to [IMPACT] via [VECTOR]. OR [COMPONENT] in [VENDOR] [PRODUCT] [VERSION] [ROOT CAUSE], which allows [ATTACKER] to [IMPACT] via [VECTOR].",
            min_length=1,
        ),
    ]


class References(RootModel[List[Reference]]):
    root: Annotated[
        List[Reference],
        Field(
            description='This is reference data in the form of URLs or file objects (uuencoded and embedded within the JSON file, exact format to be decided, e.g. we may require a compressed format so the objects require unpacking before they are "dangerous").',
            max_length=512,
            min_length=1,
        ),
    ]


class Impact(BaseModel):
    capecId: Annotated[
        Optional[str],
        Field(
            None,
            description="CAPEC ID that best relates to this impact.",
            max_length=11,
            min_length=7,
            pattern="^CAPEC-[1-9][0-9]{0,4}$",
        ),
    ]
    descriptions: Annotated[
        Descriptions,
        Field(
            description="Prose description of the impact scenario. At a minimum provide the description given by CAPEC."
        ),
    ]


class Impacts(RootModel[List[Impact]]):
    root: Annotated[
        List[Impact],
        Field(description="Collection of impacts of this vulnerability.", min_length=1),
    ]


class Scenario(BaseModel):
    lang: Language
    value: Annotated[
        str,
        Field(
            description="Description of the scenario this metrics object applies to. If no specific scenario is given, GENERAL is used as the default and applies when no more specific metric matches.",
            min_length=1,
        ),
    ]


class Configurations(RootModel[List[Description]]):
    root: Annotated[
        List[Description],
        Field(
            description="Configurations required for exploiting this vulnerability.",
            min_length=1,
        ),
    ]


class Workarounds(RootModel[List[Description]]):
    root: Annotated[
        List[Description],
        Field(
            description="Workarounds and mitigations for this vulnerability.",
            min_length=1,
        ),
    ]


class Solutions(RootModel[List[Description]]):
    root: Annotated[
        List[Description],
        Field(
            description="Information about solutions or remediations available for this vulnerability.",
            min_length=1,
        ),
    ]


class Exploits(RootModel[List[Description]]):
    root: Annotated[
        List[Description],
        Field(
            description="Information about exploits of the vulnerability.", min_length=1
        ),
    ]


class TimelineItem(BaseModel):
    time: Annotated[
        AwareDatetime,
        Field(
            description="Timestamp representing when the event in the timeline occurred. The timestamp format is based on RFC3339 and ISO ISO8601, with an optional timezone. yyyy-MM-ddTHH:mm:ssZZZZ - if the timezone offset is not given, GMT (0000) is assumed."
        ),
    ]
    lang: Annotated[
        Language,
        Field(
            description="The language used in the description of the event. The language field is included so that CVE Records can support translations. The value must be a BCP 47 language code."
        ),
    ]
    value: Annotated[
        str, Field(description="A summary of the event.", min_length=1)
    ]


class Timeline(RootModel[List[TimelineItem]]):
    root: Annotated[
        List[TimelineItem],
        Field(
            description="This is timeline information for significant events about this vulnerability or changes to the CVE Record.",
            min_length=1,
        ),
    ]


class Credit(BaseModel):
    lang: Annotated[
        Language,
        Field(
            description="The language used when describing the credits. The language field is included so that CVE Records can support translations. The value must be a BCP 47 language code."
        ),
    ]
    value: Annotated[str, Field(max_length=4096, min_length=1)]
    user: Annotated[
        Optional[UuidType],
        Field(
            None,
            description="UUID of the user being credited if present in the CVE User Registry (optional). This UUID can be used to lookup the user record in the user registry service.",
        ),
    ]
    type: Annotated[
        Optional[Type],
        Field(
            "finder",
            description="Type or role of the entity being credited (optional). finder: identifies the vulnerability.\nreporter: notifies the vendor of the vulnerability to a CNA.\nanalyst: validates the vulnerability to ensure accuracy or severity.\ncoordinator: facilitates the coordinated response process.\nremediation developer: prepares a code change or other remediation plans.\nremediation reviewer: reviews vulnerability remediation plans or code changes for effectiveness and completeness.\nremediation verifier: tests and verifies the vulnerability or its remediation.\ntool: names of tools used in vulnerability discovery or identification.\nsponsor: supports the vulnerability identification or remediation activities.",
        ),
    ]


class Credits(RootModel[List[Credit]]):
    root: Annotated[
        List[Credit],
        Field(
            description="Statements acknowledging specific people, organizations, or tools recognizing the work done in researching, discovering, remediating or helping with activities related to this CVE.",
            min_length=1,
        ),
    ]


class CnaTagsModel(RootModel[List[Union[TagExtension, CnaTags]]]):
    root: Annotated[
        List[Union[TagExtension, CnaTags]],
        Field(
            description="Tags provided by a CNA describing the CVE Record.",
            min_length=1,
        ),
    ]


class AdpTagsModel(RootModel[List[Union[TagExtension, AdpTags]]]):
    root: Annotated[
        List[Union[TagExtension, AdpTags]],
        Field(
            description="Tags provided by an ADP describing the CVE Record.",
            min_length=1,
        ),
    ]


class CnaRejectedContainer(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    providerMetadata: ProviderMetadata
    rejectedReasons: Annotated[
        Descriptions, Field(description="Reasons for rejecting this CVE Record.")
    ]
    replacedBy: Annotated[
        Optional[List[CveId]],
        Field(
            None,
            description="Contains an array of CVE IDs that this CVE ID was rejected in favor of because this CVE ID was assigned to the vulnerabilities.",
            min_length=1,
        ),
    ]


class Description1(BaseModel):
    lang: Language
    description: Annotated[
        str,
        Field(
            description="Text description of problemType, or title from CWE or OWASP.",
            min_length=1,
        ),
    ]
    cweId: Annotated[
        Optional[str],
        Field(
            None,
            description="CWE ID of the CWE that best describes this problemType entry.",
            max_length=9,
            min_length=5,
            pattern="^CWE-[1-9][0-9]*$",
        ),
    ]
    type: Annotated[
        Optional[str],
        Field(
            None,
            description="Problemtype source, text, OWASP, CWE, etc.,",
            max_length=128,
            min_length=1,
        ),
    ]
    references: Optional[References] = None


class ProblemType(BaseModel):
    descriptions: Annotated[List[Description1], Field(min_length=1)]


class ProblemTypes(RootModel[List[ProblemType]]):
    root: Annotated[
        List[ProblemType],
        Field(
            description="This is problem type information (e.g. CWE identifier). Must contain: At least one entry, can be text, OWASP, CWE, please note that while only one is required you can use more than one (or indeed all three) as long as they are correct). (CNA requirement: [PROBLEMTYPE]).",
            min_length=1,
        ),
    ]


class Metrics1(BaseModel):
    format: Annotated[
        Optional[str],
        Field(
            None,
            description="Name of the scoring format. This provides a bit of future proofing. Additional properties are not prohibited, so this will support the inclusion of proprietary formats. It also provides an easy future conversion mechanism when future score formats become part of the schema. example: cvssV44, format = 'cvssV44', other = cvssV4_4 JSON object. In the future, the other properties can be converted to score properties when they become part of the schema.",
            max_length=64,
            min_length=1,
        ),
    ]
    scenarios: Annotated[
        Optional[List[Scenario]],
        Field(
            None,
            description="Description of the scenarios this metrics object applies to. If no specific scenario is given, GENERAL is used as the default and applies when no more specific metric matches.",
            min_length=1,
        ),
    ]
    cvssV3_1: cvss_v3.Field1
    cvssV3_0: Optional[cvss_v3.Field0] = None
    cvssV2_0: Optional[cvss_v2.Field0] = None
    other: Annotated[
        Optional[Other],
        Field(
            None,
            description="A non-standard impact description, may be prose or JSON block.",
        ),
    ]


class Metrics2(BaseModel):
    format: Annotated[
        Optional[str],
        Field(
            None,
            description="Name of the scoring format. This provides a bit of future proofing. Additional properties are not prohibited, so this will support the inclusion of proprietary formats. It also provides an easy future conversion mechanism when future score formats become part of the schema. example: cvssV44, format = 'cvssV44', other = cvssV4_4 JSON object. In the future, the other properties can be converted to score properties when they become part of the schema.",
            max_length=64,
            min_length=1,
        ),
    ]
    scenarios: Annotated[
        Optional[List[Scenario]],
        Field(
            None,
            description="Description of the scenarios this metrics object applies to. If no specific scenario is given, GENERAL is used as the default and applies when no more specific metric matches.",
            min_length=1,
        ),
    ]
    cvssV3_1: Optional[cvss_v3.Field1] = None
    cvssV3_0: cvss_v3.Field0
    cvssV2_0: Optional[cvss_v2.Field0] = None
    other: Annotated[
        Optional[Other],
        Field(
            None,
            description="A non-standard impact description, may be prose or JSON block.",
        ),
    ]


class Metrics3(BaseModel):
    format: Annotated[
        Optional[str],
        Field(
            None,
            description="Name of the scoring format. This provides a bit of future proofing. Additional properties are not prohibited, so this will support the inclusion of proprietary formats. It also provides an easy future conversion mechanism when future score formats become part of the schema. example: cvssV44, format = 'cvssV44', other = cvssV4_4 JSON object. In the future, the other properties can be converted to score properties when they become part of the schema.",
            max_length=64,
            min_length=1,
        ),
    ]
    scenarios: Annotated[
        Optional[List[Scenario]],
        Field(
            None,
            description="Description of the scenarios this metrics object applies to. If no specific scenario is given, GENERAL is used as the default and applies when no more specific metric matches.",
            min_length=1,
        ),
    ]
    cvssV3_1: Optional[cvss_v3.Field1] = None
    cvssV3_0: Optional[cvss_v3.Field0] = None
    cvssV2_0: cvss_v2.Field0
    other: Annotated[
        Optional[Other],
        Field(
            None,
            description="A non-standard impact description, may be prose or JSON block.",
        ),
    ]


class Metrics4(BaseModel):
    format: Annotated[
        Optional[str],
        Field(
            None,
            description="Name of the scoring format. This provides a bit of future proofing. Additional properties are not prohibited, so this will support the inclusion of proprietary formats. It also provides an easy future conversion mechanism when future score formats become part of the schema. example: cvssV44, format = 'cvssV44', other = cvssV4_4 JSON object. In the future, the other properties can be converted to score properties when they become part of the schema.",
            max_length=64,
            min_length=1,
        ),
    ]
    scenarios: Annotated[
        Optional[List[Scenario]],
        Field(
            None,
            description="Description of the scenarios this metrics object applies to. If no specific scenario is given, GENERAL is used as the default and applies when no more specific metric matches.",
            min_length=1,
        ),
    ]
    cvssV3_1: Optional[cvss_v3.Field1] = None
    cvssV3_0: Optional[cvss_v3.Field0] = None
    cvssV2_0: Optional[cvss_v2.Field0] = None
    other: Annotated[
        Other,
        Field(
            description="A non-standard impact description, may be prose or JSON block."
        ),
    ]


class Metrics(RootModel[List[Union[Metrics1, Metrics2, Metrics3, Metrics4]]]):
    root: Annotated[
        List[Union[Metrics1, Metrics2, Metrics3, Metrics4]],
        Field(
            description="Collection of impact scores with attribution.", min_length=1
        ),
    ]


class Containers1(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    cna: CnaRejectedContainer


class CVE2(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    dataType: DataType
    dataVersion: DataVersion
    cveMetadata: CveMetadataRejected
    containers: Annotated[
        Containers1,
        Field(
            description="A set of structures (called containers) used to store vulnerability information related to a specific CVE ID provided by a specific organization participating in the CVE program. Each container includes information provided by a different source.\n\nAt minimum, a 'cna' container containing the vulnerability information provided by the CNA who initially assigned the CVE ID must be included.\n\nThere can only be one 'cna' container, as there can only be one assigning CNA."
        ),
    ]


class CnaPublishedContainer(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    providerMetadata: ProviderMetadata
    dateAssigned: Annotated[
        Optional[AwareDatetime | NaiveDatetime],
        Field(
            None,
            description="The date/time this CVE ID was associated with a vulnerability by a CNA.",
        ),
    ]
    datePublic: Annotated[
        Optional[AwareDatetime | NaiveDatetime],
        Field(
            None,
            description="If known, the date/time the vulnerability was disclosed publicly.",
        ),
    ]
    title: Annotated[
        Optional[str],
        Field(
            None,
            description="A title, headline, or a brief phrase summarizing the CVE record. Eg., Buffer overflow in Example Soft.",
            max_length=256,
            min_length=1,
        ),
    ]
    descriptions: Descriptions
    affected: Affected
    problemTypes: Optional[ProblemTypes] = None
    references: Optional[References] = None
    impacts: Optional[Impacts] = None
    metrics: Optional[Metrics] = None
    configurations: Optional[Configurations] = None
    workarounds: Optional[Workarounds] = None
    solutions: Optional[Solutions] = None
    exploits: Optional[Exploits] = None
    timeline: Optional[Timeline] = None
    credits: Optional[Credits] = None
    source: Optional[Source] = None
    tags: Optional[CnaTagsModel] = None
    taxonomyMappings: Optional[TaxonomyMappings] = None


CnaPublishedContainer.model_rebuild()


class AdpContainer(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    providerMetadata: ProviderMetadata
    datePublic: Annotated[
        Optional[AwareDatetime | NaiveDatetime],
        Field(
            None,
            description="If known, the date/time the vulnerability was disclosed publicly.",
        ),
    ]
    title: Annotated[
        Optional[str],
        Field(
            None,
            description="A title, headline, or a brief phrase summarizing the information in an ADP container.",
            max_length=256,
            min_length=1,
        ),
    ]
    descriptions: Optional[Descriptions] = None
    affected: Optional[Affected] = None
    problemTypes: Optional[ProblemTypes] = None
    references: Optional[References] = None
    impacts: Optional[Impacts] = None
    metrics: Optional[Metrics] = None
    configurations: Optional[Configurations] = None
    workarounds: Optional[Workarounds] = None
    solutions: Optional[Solutions] = None
    exploits: Optional[Exploits] = None
    timeline: Optional[Timeline] = None
    credits: Optional[Credits] = None
    source: Optional[Source] = None
    tags: Optional[AdpTagsModel] = None
    taxonomyMappings: Optional[TaxonomyMappings] = None


class Containers(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    cna: Optional[CnaPublishedContainer]
    adp: Annotated[Optional[List[AdpContainer]], Field(None, min_length=1)]


class CVE1(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    dataType: DataType
    dataVersion: DataVersion
    cveMetadata: CveMetadataPublished
    containers: Annotated[
        Containers,
        Field(
            description="A set of structures (called containers) used to store vulnerability information related to a specific CVE ID provided by a specific organization participating in the CVE program. Each container includes information provided by a different source.\n\nAt a minimum, a 'cna' container containing the vulnerability information provided by the CNA who initially assigned the CVE ID must be included.\n\nThere can only be one 'cna' container, as there can only be one assigning CNA. However, there can be multiple 'adp' containers, allowing multiple organizations participating in the CVE program to add additional information related to the vulnerability. For the most part, the 'cna' and 'adp' containers contain the same properties. The main differences are the source of the information. The 'cna' container requires the CNA to include certain fields, while the 'adp' container does not."
        ),
    ]


class CVE(RootModel[Union[CVE1, CVE2]]):
    root: Annotated[
        Union[CVE1, CVE2],
        Field(
            description="cve-schema specifies the CVE JSON record format. This is the blueprint for a rich set of JSON data that can be submitted by CVE Numbering Authorities (CNAs) and Authorized Data Publishers (ADPs) to describe a CVE Record. Some examples of CVE Record data include CVE ID number, affected product(s), affected version(s), and public references. While those specific items are required when assigning a CVE, there are many other optional data in the schema that can be used to enrich CVE Records for community benefit. Learn more about the CVE program at [the official website](https://cve.mitre.org). This CVE JSON record format is defined using JSON Schema. Learn more about JSON Schema [here](https://json-schema.org/).",
            title="CVE JSON record format",
        ),
    ]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import base64
import logging
import os
import shutil
import types

from rich.console import Console
from rich.live import Live
from rich.markdown import Markdown
from rich.table import Table

from vdb.lib import config, db6 as db_lib, search
from vdb.lib.aqua import AquaSource
from vdb.lib.cve_model import CVE
from vdb.lib.gha import GitHubSource
from vdb.lib.osv import OSVSource

ORAS_AVAILABLE = False
# oras is an optional dependency
try:
    from vdb.lib.orasclient import download_image

    ORAS_AVAILABLE = True
except ImportError:
    pass

console = Console()

logging.basicConfig(
    level=logging.INFO, format="%(levelname)s [%(asctime)s] %(message)s"
)
LOG = logging.getLogger(__name__)
for _ in ("httpx",):
    logging.getLogger(_).disabled = True

AT_LOGO = r"""
             ___
  /\  ._  ._  | |_  ._ _   _. _|_
 /--\ |_) |_) | | | | (/_ (_|  |_
      |   |
"""


def build_args():
    """
    Constructs command line arguments for the vulnerability-db tool
    """
    parser = argparse.ArgumentParser(
        description="AppThreat's vulnerability database and package search library with a built-in sqlite based storage."
    )
    parser.add_argument(
        "--clean",
        action="store_true",
        default=False,
        dest="clean",
        help="Clear the vulnerability database cache from platform specific user_data_dir.",
    )
    parser.add_argument(
        "--cache",
        action="store_true",
        default=False,
        dest="cache",
        help="Cache vulnerability information in platform specific user_data_dir.",
    )
    parser.add_argument(
        "--cache-os",
        action="store_true",
        default=False,
        dest="cache_os",
        help="Cache OS vulnerability information in platform specific user_data_dir.",
    )
    parser.add_argument(
        "--only-osv",
        action="store_true",
        default=False,
        dest="only_osv",
        help="Use only OSV as the source. Use with --cache.",
    )
    parser.add_argument(
        "--only-aqua",
        action="store_true",
        default=False,
        dest="only_aqua",
        help="Use only Aqua vuln-list as the source. Use with --cache.",
    )
    parser.add_argument(
        "--only-ghsa",
        action="store_true",
        default=False,
        dest="only_ghsa",
        help="Use only recent ghsa as the source. Use with --cache.",
    )
    parser.add_argument(
        "--search",
        dest="search",
        help="Search for the package or CVE ID in the database. Use purl, cpe, or git http url.",
    )
    parser.add_argument(
        "--list-malware",
        action="store_true",
        default=False,
        dest="list_malware",
        help="List latest malwares with CVE ID beginning with MAL-.",
    )
    parser.add_argument(
        "--bom",
        dest="bom_file",
        help="Search for packages in the CycloneDX BOM file.",
    )
    parser.add_argument(
        "--download-image",
        action="store_true",
        default=False,
        dest="download_image",
        help="Downloaded pre-created vdb image to platform specific user_data_dir.",
    )
    return parser.parse_args()


def add_table_row(table: Table, res: dict, added_row_keys: dict):
    # matched_by is the purl or cpe string
    row_key = f"""{res["matched_by"]}|{res.get("source_data_hash")}"""
    # Filter duplicate rows from getting printed
    if added_row_keys.get(row_key):
        return
    source_data: CVE = res.get("source_data")
    descriptions = []
    if (
        source_data.root.containers.cna
        and source_data.root.containers.cna.descriptions
        and source_data.root.containers.cna.descriptions.root
    ):
        for adesc in source_data.root.containers.cna.descriptions.root:
            description = (
                "\n".join(
                    [
                        base64.b64decode(sm.value).decode("utf-8")
                        for sm in adesc.supportingMedia
                    ]
                )
                if adesc.supportingMedia
                else adesc.value
            )
            description = description.replace("\\n", "\n").replace("\\t", "  ")
            descriptions.append(description)
    table.add_row(
        res.get("cve_id"),
        res.get("matched_by"),
        Markdown("\n".join(descriptions), justify="left", hyperlinks=True),
    )
    added_row_keys[row_key] = True


def print_results(results):
    added_row_keys = {}
    table = Table(title="VDB Results", show_lines=True)
    table.add_column("CVE", justify="left")
    table.add_column("Locator")
    table.add_column("Description")
    if isinstance(results, types.GeneratorType):
        with Live(
            table, console=console, refresh_per_second=4, vertical_overflow="visible"
        ):
            for result_gen in results:
                if isinstance(result_gen, dict):
                    add_table_row(table, result_gen, added_row_keys)
                if isinstance(result_gen, types.GeneratorType):
                    for res in result_gen:
                        add_table_row(table, res, added_row_keys)
    elif isinstance(results, list):
        for res in results:
            add_table_row(table, res, added_row_keys)
        console.print(table)


def main():
    """Main function"""
    args = build_args()
    print(AT_LOGO)
    if args.clean:
        if os.path.exists(config.DATA_DIR):
            shutil.rmtree(config.DATA_DIR, ignore_errors=True)
    if args.download_image:
        if ORAS_AVAILABLE:
            LOG.info(
                "Downloading vdb image from %s to %s",
                config.VDB_DATABASE_URL,
                config.DATA_DIR,
            )
            download_image(config.VDB_DATABASE_URL, config.DATA_DIR)
        else:
            console.print(
                "Oras library is not available. Install using pip install appthreat-vulnerability-db[oras] and then re-run this command."
            )
    elif args.cache or args.cache_os:
        db_lib.get()
        db_lib.clear_all()
        if args.only_osv:
            sources = [OSVSource()]
        elif args.only_aqua:
            sources = [AquaSource()]
        elif args.only_ghsa:
            sources = [GitHubSource()]
        else:
            sources = [OSVSource(), GitHubSource()]
        # AquaSource also includes NVD
        if args.cache_os:
            sources.insert(0, AquaSource())
        for s in sources:
            LOG.info("Refreshing %s", s.__class__.__name__)
            s.refresh()
        cve_data_count, cve_index_count = db_lib.stats()
        console.print(
            "cve_data_count", cve_data_count, "cve_index_count", cve_index_count
        )
        db_lib.optimize_and_close_all()
    if args.search:
        results = search.search_by_any(args.search, with_data=True)
        if results:
            print_results(results)
        else:
            console.print("No results found!")
    elif args.bom_file:
        if os.path.exists(args.bom_file):
            results_generator = search.search_by_cdx_bom(args.bom_file, with_data=True)
            print_results(results_generator)
    elif args.list_malware:
        results_generator = search.latest_malware(with_data=True)
        print_results(results_generator)


if __name__ == "__main__":
    main()



import os
import sys

import apsw

from vdb.lib import config

db_conn: apsw.Connection = None
index_conn: apsw.Connection = None
tables_created = False
DB_FILE_SEP = "///" if sys.platform == "win32" else "//"


def ensure_schemas(db_conn_obj: apsw.Connection, index_conn_obj: apsw.Connection):
    """Create the sqlite tables and indexes in case they don't exist"""
    db_conn_obj.execute(
        "CREATE TABLE if not exists cve_data(cve_id TEXT NOT NULL, type TEXT NOT NULL, namespace TEXT, name TEXT NOT NULL, source_data BLOB NOT NULL, override_data BLOB, source_data_hash TEXT NOT NULL, purl_prefix TEXT NOT NULL);")
    db_conn_obj.pragma("synchronous", "OFF")
    db_conn_obj.pragma("journal_mode", "MEMORY")
    index_conn_obj.execute(
        "CREATE TABLE if not exists cve_index(cve_id TEXT NOT NULL, type TEXT NOT NULL, namespace TEXT, name TEXT NOT NULL, vers TEXT NOT NULL, purl_prefix TEXT NOT NULL);")
    index_conn_obj.pragma("synchronous", "OFF")
    index_conn_obj.pragma("journal_mode", "MEMORY")


def get(db_file: str = config.VDB_BIN_FILE, index_file: str = config.VDB_BIN_INDEX, read_only=False) -> (
        apsw.Connection, apsw.Connection):
    """Gets the connection to the index and the data databases. Raises apsw.CantOpenError if the database is not available."""
    global db_conn, index_conn, tables_created
    if not db_file.startswith("file:"):
        db_file = f"file:{DB_FILE_SEP}{os.path.abspath(db_file)}"
    if not index_file.startswith("file:"):
        index_file = f"file:{DB_FILE_SEP}{os.path.abspath(index_file)}"
    rw_flags = apsw.SQLITE_OPEN_URI | apsw.SQLITE_OPEN_NOFOLLOW | apsw.SQLITE_OPEN_CREATE | apsw.SQLITE_OPEN_READWRITE
    ro_flags = apsw.SQLITE_OPEN_URI | apsw.SQLITE_OPEN_NOFOLLOW | apsw.SQLITE_OPEN_READONLY
    # To avoid apsw.CantOpenError, we support readonly mode only if the db exists
    flags = ro_flags if read_only and os.path.exists(db_file) and os.path.exists(index_file) else rw_flags
    if not db_conn:
        db_conn = apsw.Connection(db_file, flags=flags)
    if not index_conn:
        index_conn = apsw.Connection(index_file, flags=flags)
    if not tables_created:
        ensure_schemas(db_conn, index_conn)
        tables_created = True
    return db_conn, index_conn


def stats():
    cve_data_count = 0
    res = db_conn.execute("SELECT count(*) FROM cve_data").fetchone()
    if res:
        cve_data_count = res[0]
    cve_index_count = 0
    res = index_conn.execute("SELECT count(*) FROM cve_index").fetchone()
    if res:
        cve_index_count = res[0]
    return cve_data_count, cve_index_count


def clear_all():
    if db_conn:
        db_conn.execute("DELETE FROM cve_data;")
    if index_conn:
        index_conn.execute("DELETE FROM cve_index;")


def optimize_and_close_all():
    """
    Safely close the connections by creating indexes and vacuuming if needed.
    """
    if db_conn:
        db_conn.execute(
            "CREATE INDEX if not exists idx1 on cve_data(cve_id, purl_prefix);")
        db_conn.execute("VACUUM;")
        db_conn.close()
    if index_conn:
        index_conn.execute(
            "CREATE INDEX if not exists cidx1 on cve_index(cve_id);")
        index_conn.execute(
            "CREATE INDEX if not exists cidx2 on cve_index(type, namespace, name);")
        index_conn.execute(
            "CREATE INDEX if not exists cidx3 on cve_index(type, name);")
        index_conn.execute(
            "CREATE INDEX if not exists cidx4 on cve_index(namespace, name);")
        index_conn.execute(
            "CREATE INDEX if not exists cidx5 on cve_index(purl_prefix);")
        index_conn.execute("VACUUM;")
        index_conn.close()



# pylint: disable=C0115, C0103, C0301

from __future__ import annotations

from enum import Enum
from typing import Annotated, Optional

from pydantic import BaseModel, Field

from vdb.lib.cve_model.common import (
    AttackComplexityType,
    AttackVectorType,
    CiaRequirementType,
    CiaType,
    ConfidenceType,
    ExploitCodeMaturityType,
    ModifiedAttackComplexityType,
    ModifiedAttackVectorType,
    ModifiedCiaType,
    ModifiedPrivilegesRequiredType,
    ModifiedScopeType,
    ModifiedUserInteractionType,
    PrivilegesRequiredType,
    RemediationLevelType,
    ScopeType,
    ScoreType,
    SeverityType,
    UserInteractionType,
)


class Version(Enum):
    field_3_1 = "3.1"


class Version1Model(Enum):
    field_3_0 = "3.0"


class Field1(BaseModel):
    version: Annotated[Version, Field(description="CVSS Version")]
    vectorString: Annotated[
        str,
        Field(
            pattern="^CVSS:3[.]1/((AV:[NALP]|AC:[LH]|PR:[NLH]|UI:[NR]|S:[UC]|[CIA]:[NLH]|E:[XUPFH]|RL:[XOTWU]|RC:[XURC]|[CIA]R:[XLMH]|MAV:[XNALP]|MAC:[XLH]|MPR:[XNLH]|MUI:[XNR]|MS:[XUC]|M[CIA]:[XNLH])/)*(AV:[NALP]|AC:[LH]|PR:[NLH]|UI:[NR]|S:[UC]|[CIA]:[NLH]|E:[XUPFH]|RL:[XOTWU]|RC:[XURC]|[CIA]R:[XLMH]|MAV:[XNALP]|MAC:[XLH]|MPR:[XNLH]|MUI:[XNR]|MS:[XUC]|M[CIA]:[XNLH])$"
        ),
    ]
    attackVector: Optional[AttackVectorType] = None
    attackComplexity: Optional[AttackComplexityType] = None
    privilegesRequired: Optional[PrivilegesRequiredType] = None
    userInteraction: Optional[UserInteractionType] = None
    scope: Optional[ScopeType] = None
    confidentialityImpact: Optional[CiaType] = None
    integrityImpact: Optional[CiaType] = None
    availabilityImpact: Optional[CiaType] = None
    baseScore: ScoreType
    baseSeverity: SeverityType
    exploitCodeMaturity: Optional[ExploitCodeMaturityType] = None
    remediationLevel: Optional[RemediationLevelType] = None
    reportConfidence: Optional[ConfidenceType] = None
    temporalScore: Optional[ScoreType] = None
    temporalSeverity: Optional[SeverityType] = None
    confidentialityRequirement: Optional[CiaRequirementType] = None
    integrityRequirement: Optional[CiaRequirementType] = None
    availabilityRequirement: Optional[CiaRequirementType] = None
    modifiedAttackVector: Optional[ModifiedAttackVectorType] = None
    modifiedAttackComplexity: Optional[ModifiedAttackComplexityType] = None
    modifiedPrivilegesRequired: Optional[ModifiedPrivilegesRequiredType] = None
    modifiedUserInteraction: Optional[ModifiedUserInteractionType] = None
    modifiedScope: Optional[ModifiedScopeType] = None
    modifiedConfidentialityImpact: Optional[ModifiedCiaType] = None
    modifiedIntegrityImpact: Optional[ModifiedCiaType] = None
    modifiedAvailabilityImpact: Optional[ModifiedCiaType] = None
    environmentalScore: Optional[ScoreType] = None
    environmentalSeverity: Optional[SeverityType] = None


class Field0(BaseModel):
    version: Annotated[Version1Model, Field(description="CVSS Version")]
    vectorString: Annotated[
        str,
        Field(
            pattern="^CVSS:3[.]0/((AV:[NALP]|AC:[LH]|PR:[UNLH]|UI:[NR]|S:[UC]|[CIA]:[NLH]|E:[XUPFH]|RL:[XOTWU]|RC:[XURC]|[CIA]R:[XLMH]|MAV:[XNALP]|MAC:[XLH]|MPR:[XUNLH]|MUI:[XNR]|MS:[XUC]|M[CIA]:[XNLH])/)*(AV:[NALP]|AC:[LH]|PR:[UNLH]|UI:[NR]|S:[UC]|[CIA]:[NLH]|E:[XUPFH]|RL:[XOTWU]|RC:[XURC]|[CIA]R:[XLMH]|MAV:[XNALP]|MAC:[XLH]|MPR:[XUNLH]|MUI:[XNR]|MS:[XUC]|M[CIA]:[XNLH])$"
        ),
    ]
    attackVector: Optional[AttackVectorType] = None
    attackComplexity: Optional[AttackComplexityType] = None
    privilegesRequired: Optional[PrivilegesRequiredType] = None
    userInteraction: Optional[UserInteractionType] = None
    scope: Optional[ScopeType] = None
    confidentialityImpact: Optional[CiaType] = None
    integrityImpact: Optional[CiaType] = None
    availabilityImpact: Optional[CiaType] = None
    baseScore: ScoreType
    baseSeverity: SeverityType
    exploitCodeMaturity: Optional[ExploitCodeMaturityType] = None
    remediationLevel: Optional[RemediationLevelType] = None
    reportConfidence: Optional[ConfidenceType] = None
    temporalScore: Optional[ScoreType] = None
    temporalSeverity: Optional[SeverityType] = None
    confidentialityRequirement: Optional[CiaRequirementType] = None
    integrityRequirement: Optional[CiaRequirementType] = None
    availabilityRequirement: Optional[CiaRequirementType] = None
    modifiedAttackVector: Optional[ModifiedAttackVectorType] = None
    modifiedAttackComplexity: Optional[ModifiedAttackComplexityType] = None
    modifiedPrivilegesRequired: Optional[ModifiedPrivilegesRequiredType] = None
    modifiedUserInteraction: Optional[ModifiedUserInteractionType] = None
    modifiedScope: Optional[ModifiedScopeType] = None
    modifiedConfidentialityImpact: Optional[ModifiedCiaType] = None
    modifiedIntegrityImpact: Optional[ModifiedCiaType] = None
    modifiedAvailabilityImpact: Optional[ModifiedCiaType] = None
    environmentalScore: Optional[ScoreType] = None
    environmentalSeverity: Optional[SeverityType] = None
